# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Defines `BLProp`, a high-level wrapper for interacting with Blender properties."""

import functools
import typing as typ
import uuid

import bpy
import pydantic as pyd

from blroots.services import bl_prop_cache

from .signal import Signal

SupportedBLPropType: typ.TypeAlias = (
	bpy.types.ID | bpy.types.Operator | bpy.types.AddonPreferences
)


####################
# - Blender Property (Abstraction)
####################
class BLProp(pyd.BaseModel, frozen=True):
	"""Convenient, fast, and safe interface to a property on a Blender class.

	To define properties on Blender classes, one typically uses `bpy.props`.
	For more demanding applications, however, this direct interface can be very frustrating to work with.
	However, there are many drastic caveats to directly using this interface:

	- **Flexibility**: Only `bool`, `int`, `float`, and `str` can be used, with minor variants thereof.
		Representing anything else requires parsing, resulting either in widespread bulky and error-prone boilerplate.
		Alternatively, one may simply limit extension functionality to what is easily represented with `bpy.props`.
	- **Thread Safety**: Properties are absolutely not thread-safe.
		For certain applications, the ability to handle properties in the background is incredibly important for maintaining a smooth user interface.
	- **Real-World Types**: Simple lists and dict-like structures can quickly become overwhelming to work with, let alone something like a `dataclass` / `pydantic.BaseModel`.
		One must declare and register `PropertyGroup`s, manage specialized `draw()` code at each point of use, and most importantly, constantly manage translating traditional data-structures to these specialized representations.
		This must be done even for the simpest of lists.
	- **Consistency and Stability**: One must often be very careful when working with `bpy.props`.
		Something as simple as a drop-down menu with dynamic elements can crash Blender without warning or reason.
		Use of `get`, `set`, or `update` launches threaded contexts, which theoretically prohibits use of `bpy` - but in practice is often OK, leading many users to unintentionally create race conditions.

	By treating `bpy.props` as a low-level interface, `BLProp` solves each and every one of these issues by providing an utterly friendly backend to ex. `BLField`.

	By using `BLProp`, via ex. `BLField`, the ability of extensions to quickly provide concise, powerful, and reliable functionality is greatly enhanced.

	Attributes:
		type_hint: The type hint constraining the identity of the
		name: The name of the Blender property, as one uses it.
		display_name: The name of the Blender property, as one uses it.
		prop_info: Specifies the property's particular behavior, including subtype and UI.
		prop_type: The type to associate with the property.
			Especially relevant for structured deserialization.
		bl_prop_type: Identifier encapsulating which Blender property used for data storage, and how.
	"""

	type_hint: type

	name: str
	display_name: str
	description: str

	default_value: typ.Any

	# TODO: Support certain property flags universally.

	####################
	# - Computed
	####################
	@functools.cached_property
	def bl_name(self):
		"""Deduces the internal attribute name where the raw Blender property data will be persisted."""
		return f'blprop__{self.name}'

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Cheaply deduce whether this BLProp implements the given type."""
		raise NotImplementedError

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.FloatProperty]:
		"""The underlying `bpy.props` class backing this property."""
		raise NotImplementedError

	@property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		raise NotImplementedError

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> typ.Any:
		"""Encode a value for compatibility with this Blender property, using the encapsulated types."""
		raise NotImplementedError

	def decode(self, value: typ.Any) -> typ.Any:
		"""Encode a value for compatibility with this Blender property, using the encapsulated types."""
		raise NotImplementedError

	def handle_signal(self, signal: Signal) -> None:
		"""Handle a signal in a manner specific to a Blender property."""
		raise NotImplementedError

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: SupportedBLPropType,
		layout: bpy.types.UILayout,
	) -> None:
		"""Draw this property to the given Blender `UILayout`.

		Notes:
			Can and should be overridden by subclasses.
			Consider this a sensible default.
		"""
		## TODO:
		## - Universal support for icons?
		layout.prop(bl_instance, self.bl_name, text=self.display_name)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""The priority with which a type hint will resolve to this implementation in case of a conflict.

		Lower implies higher priority.
		"""
		raise NotImplementedError

	####################
	# - Creation
	####################
	@classmethod
	def from_type_hint(
		cls,
		field_type_hint: type,
		*,
		name: str,
		default_value: typ.Any,
		display_name: str,
		description: str,
		**bl_prop_info: dict[str, typ.Any],
	) -> typ.Self:
		# Deduce Subclass from Type Check
		subclasses = sorted(
			cls.__subclasses__(),
			key=lambda subclass: subclass.priority(),
		)
		supported_subclasses = [
			subclass
			for subclass in subclasses
			if subclass.supports_type_hint(field_type_hint)
		]
		if supported_subclasses:
			BLPropSubclass = supported_subclasses[0]

			# Construct the Subclass
			return BLPropSubclass(
				type_hint=field_type_hint,
				name=name,
				display_name=display_name,
				description=description,
				default_value=default_value,
				**bl_prop_info,
			)
		msg = f"Tried to construct BLProp (name='{name}') from type hint '{field_type_hint}', but no implementing subclass was found to dispatch to."
		raise ValueError(msg)

	####################
	# - Initialization
	####################
	@property
	def bpy_prop(self) -> bpy.types.Property:
		"""Fully initialized `bpy.props` object."""
		return self.bpy_prop_type(
			name=self.display_name,
			description=self.description,
			update=lambda bl_instance, _: self.invalidate_nonpersist(bl_instance),
			**self.bpy_prop_kwargs,
		)

	def init_on_bl_class(
		self,
		bl_class: type[SupportedBLPropType],
	) -> None:
		"""Declare the Blender property on a Blender class, ensuring that the property will be available to all `SupportedBLPropType` respecting instances of that class.

		Parameters:
			obj_type: The exact object type that will be stored in the Blender property.
				**Must** be chosen such that `BLPropType.from_type(obj_type) == self`.
		"""

		# Add Method: Reset Instance ID
		def reset_instance_id(bl_instance: SupportedBLPropType) -> None:
			bl_instance.instance_id = str(uuid.uuid4())

		bl_class.reset_instance_id = reset_instance_id

		# ID: Dynamic Assignment to Class Variable
		if issubclass(bl_class, bpy.types.ID):
			# Set Instance ID
			if not hasattr(bl_class, 'instance_id'):
				bl_class.instance_id = bpy.props.StringProperty()

			# Set Property
			setattr(bl_class, self.bl_name, self.bpy_prop)

		# bpy_struct: Static Assignment to Class Annotations
		elif issubclass(
			bl_class,
			bpy.types.AddonPreferences
			| bpy.types.Operator
			| bpy.types.Node
			| bpy.types.NodeSocket,
		):
			# Set Instance ID
			if 'instance_id' not in bl_class.__annotations__:
				bl_class.__annotations__['instance_id'] = bpy.props.StringProperty()

			# Set Annotation
			bl_class.__annotations__[self.bl_name] = self.bpy_prop

	####################
	# - Instance Methods
	####################
	def read_nonpersist(self, bl_instance: SupportedBLPropType | None) -> typ.Any:
		"""Read the non-persistent cache value for this property.

		Notes:
			**Never reads cached BLPointers**; invokes `self.read()` instead.
			`BLPointer`s must align perfectly with Blender's internal logic, and as such, the cache cannot get involved, else we risk access-after-free crashes.

		Returns:
			Generally, the cache value, with two exceptions.

			- `Signal.CacheNotReady`: When either `bl_instance` is None, or it doesn't yet have a unique `bl_instance.instance_id`.
				Indicates that the instance is not yet ready for use.
				For nodes, `init()` has not yet run.
				For sockets, `preinit()` has not yet run.

			- `Signal.CacheEmpty`: When the cache has no entry.
				A good idea might be to fill it immediately with `self.write_nonpersist(bl_instance)`.
		"""
		value = bl_prop_cache.read(
			bl_instance,
			key=self.bl_name,
			use_nonpersist=True,
			use_persist=False,
		)

		# if value is Signal.CacheNotReady:
		# msg = f"Tried to read value '{self.bl_name}' from non-persistent cache of bl_instance '{bl_instance}', but the cache was not yet ready."
		# raise ValueError(msg)

		# if value is Signal.CacheEmpty:
		# msg = f"Tried to read value '{self.bl_name}' from non-persistent cache of bl_instance '{bl_instance}', but the cache is empty."
		# raise ValueError(msg)

		return value

	def read_persist(self, bl_instance: SupportedBLPropType) -> typ.Any:
		"""Read the persistent cache value for this property.

		Parameters:
			bl_instance: The Blender object to
				**NOTE**: `bl_instance` must not be `None`, as neighboring methods sometimes allow.
		"""
		value = self.decode(
			bl_prop_cache.read(
				bl_instance,
				key=self.bl_name,
				use_nonpersist=False,
				use_persist=True,
			)
		)

		if value is Signal.CacheNotReady:
			msg = f"Tried to read value '{self.bl_name}' from persistent cache of bl_instance '{bl_instance}', but the cache was not yet ready."
			raise ValueError(msg)

		if value is Signal.CacheEmpty:
			msg = f"Tried to read value '{self.bl_name}' from persistent cache of bl_instance '{bl_instance}', but the cache is empty."
			raise ValueError(msg)

		return value

	def write_nonpersist(
		self, bl_instance: SupportedBLPropType, value: typ.Any
	) -> None:
		"""Write the non-persistent cache value for this property.

		Parameters:
			bl_instance: The Blender object to
				**NOTE**: `bl_instance` must not be `None`, as neighboring methods sometimes allow.
		"""
		signal = bl_prop_cache.write(
			bl_instance,
			key=self.bl_name,
			value=value,
			use_nonpersist=True,
			use_persist=False,
		)

		# if signal is Signal.CacheNotReady:
		# msg = f"Tried to write value '{self.bl_name}={value}' to non-persistent cache of bl_instance '{bl_instance}', but the cache was not yet ready."
		# raise ValueError(msg)

	def write_persist(self, bl_instance: SupportedBLPropType, value: typ.Any) -> None:
		"""Write the persistent cache value for this property.

		Parameters:
			bl_instance: The Blender object to
				**NOTE**: `bl_instance` must not be `None`, as neighboring methods sometimes allow.
		"""
		signal = bl_prop_cache.write(
			bl_instance,
			key=self.bl_name,
			value=self.encode(value),
			use_nonpersist=False,
			use_persist=True,
		)

		if signal is Signal.CacheNotReady:
			msg = f"Tried to write value '{self.bl_name}={value}' to persistent cache of bl_instance '{bl_instance}', but the cache was not yet ready."
			raise ValueError(msg)

		self.write_nonpersist(bl_instance, value)

	def invalidate_nonpersist(self, bl_instance: SupportedBLPropType | None) -> None:
		"""Invalidate the non-persistent cache value for this property.

		Parameters:
			bl_instance: The Blender object to
				**NOTE**: `bl_instance` must not be `None`, as neighboring methods sometimes allow.
		"""
		bl_prop_cache.invalidate_nonpersist(
			bl_instance,
			self.bl_name,
		)
