# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements `ClassProp`."""

import functools
import inspect
import typing as typ

import bpy

from blroots.utils import bl_instance, logger

from ..bl_prop import BLProp

log = logger.get(__name__)

PRIORITY = 11


class ClassProp(BLProp, frozen=True):
	"""A single constrained floating-point number.

	Parameters:
		abs_min: The absolute minimum value of the floating point number.
		abs_max: The absolute maximum value of the floating point number.
		soft_min: The lowest value that can be set using the UI slider.
		soft_max: The highest value that can be set using the UI slider.
		float_prec: The number of decimal places displayed in the UI.
		float_step: The size of each mouse-move increment when using the UI slider.
			_Divide this integer by `100` for actual change per increment._
	"""

	default_value: None

	####################
	# - Type Support
	####################
	@classmethod
	def supports_type_hint(cls, field_type_hint: type) -> bool:
		"""Support only the explicit type `float` aka. `builtins.float`."""
		return inspect.isclass(field_type_hint) and hasattr(
			field_type_hint, 'property_group'
		)

	@property
	def bpy_property_group(self) -> type[bpy.props.PropertyGroup]:
		"""Dynamically create a PropertyGroup."""
		## TODO: We can make it; the trouble is, who registers it, and when?
		## - Also, there's an underlying class instance
		pass

	####################
	# - bpy.props Interface
	####################
	@property
	def bpy_prop_type(self) -> type[bpy.props.PointerProperty]:
		"""Use `bpy.props.PointerProperty` to reference a particular pre-made `bpy.types.PropertyGroup` subclass."""
		return bpy.props.PointerProperty  # type: ignore[no-any-return]

	@functools.cached_property
	def bpy_prop_kwargs(self) -> dict[str, typ.Any]:
		"""Keyword arguments to pass to `self.bpy_prop_type`."""
		return {
			'type': self.bpy_property_group,
		}

	####################
	# - Encode/Decode
	####################
	def encode(self, value: typ.Any) -> float:
		"""Encode a value that supports `float()` to a Blender float."""
		if isinstance(value, self.type_hint):
			# TODO: Dump the model to a dictionary? How to handle the nested BLProps? Better to delegate somehow to the PropertyGroup?
			pass

		raise NotImplementedError

	def decode(self, raw_bl_value: float) -> float:
		"""Decode a Blender float by doing nothing."""
		return raw_bl_value

	####################
	# - UI
	####################
	def draw(
		self,
		bl_instance: bl_instance.BLInstance,
		layout: bpy.types.UILayout,
		_: bpy.types.Context,
	) -> None:
		"""Draw this property to the given Blender `UILayout`."""
		prop_group = getattr(bl_instance, self.bl_name)
		self.type_hint.__class__.draw(prop_group)

	####################
	# - Selection
	####################
	@classmethod
	def priority(cls) -> int:
		"""Priority with which to prefer this implementation over others compatible with the same type hint."""
		return PRIORITY
