# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A visual DSL for electromagnetic simulation design and analysis implemented as a Blender node editor."""

import bpy

from blroots import registration
from blroots import typ as bltyp
from blroots.utils import bl_logger
from blroots.utils.bl_handlers import BLHandlers
from blroots.utils.bl_keymap_item import BLKeymapItem

log = bl_logger.get(__name__)


####################
# - Registration
####################
def register(
	blext_pkg: bltyp.BLExtPkgName,
	*,
	classes_prefs: list[type[bltyp.BLClass]],
	classes: list[type[bltyp.BLClass]],
	handlers: BLHandlers,
	keymap_items: list[BLKeymapItem],
) -> None:
	"""Implements extension registration as a standardized procedure.

	1. Extension preferences classes are registered first. _This ensure that these values are available to subsequently registered Blender classes._
	2. Extension loggers are updated to reflect the extension preferences.
	3. The extension's classes are registered.
	4. The extension's handlers are registered.
	5. The extension's keymaps are registered.

	Notes:
		Extensions may use this function to greatly simplify their `register()` function.

	Parameters:
		blext_pkg: The full extension package name `bl_ext.<repo>.<name>`.
			From `__init__.py`, this should be set to `__package__`.
		classes_preferences: Classes associated with extension preferences.
			After registration of these classes, `bpy.context.preferences.addons.get(...)` should return non-`None` for this addon.
		classes_preferences: Classes associated with extension preferences.

	Raises:
		RuntimeError: If addon preferences fail to register.
	"""
	log.info('Registering Extension Preferences: %s', blext_pkg)
	registration.register_classes(blext_pkg, classes_prefs)

	# Access Addon Preferences
	## - It may not be
	addon = bpy.context.preferences.addons.get(blext_pkg)
	if addon is not None and (addon_prefs := addon.preferences) is not None:
		# Addon Prefs [bl_logger]: Update Loggers
		# - Update existing loggers to use the settings defined in the user preferences.
		if getattr(addon_prefs, 'enable_integration__bl_logger', False):
			addon_prefs.conform_all_loggers_to_prefs()

		log.info('Registering Extension: %s', blext_pkg)

		# Perform Registrations
		registration.register_classes(blext_pkg, classes)
		registration.register_handlers(blext_pkg, handlers)
		registration.register_keymaps(blext_pkg, keymap_items)

		log.info('Finished Registration of Extension: %s', blext_pkg)
	else:
		msg = f"Addon preferences did not register for extension '{blext_pkg}'. Aborting addon registration."
		raise RuntimeError(msg)


def unregister(
	blext_pkg: bltyp.BLExtPkgName,
) -> None:
	"""Unregisters anything that was registered by the addon.

	Notes:
		Run by Blender when disabling and/or uninstalling the addon.

		This doesn't clean `sys.modules`.
		We rely on the hope that Blender has extension-extension module isolation.
	"""
	log.info('Unregistering Extension: %s', blext_pkg)

	registration.unregister_keymaps(blext_pkg)
	registration.unregister_handlers(blext_pkg)
	registration.unregister_classes(blext_pkg)

	log.info('Finished Unregistering Extension: %s', blext_pkg)
