import logging
import typing as typ
from pathlib import Path

import bpy

from blroots import gather
from blroots.bl_prop import BLField
from blroots.utils import bl_logger


####################
# - User Preferences Mixin
####################
def LoggingPrefs(blext_pkg: str) -> type[type]:
	class _LoggingPrefs(bpy.types.AddonPreferences):
		"""Implements the `bl_logger` user preferences integration.

		Notes:
			To check for the presence of this integration, use:
			```python
			getattr(addon_prefs, 'enable_integration__bl_logger', False)
			```
		"""

		enable_integration__bl_logger: bool = True

		####################
		# - Properties
		####################
		# Logging
		## File Logging
		use_log_file: bool = BLField(
			gather.blext_info(blext_pkg).use_log_file,
			display_name='Log to File',
		)
		# log_file_level: bl_logger.BLExtLogLevel = BLField(
		# gather.blext_info(blext_pkg).log_file_level,
		# display_name='File Log Level',
		# )
		# log_file_path: Path = BLField(
		# gather.blext_info(blext_pkg).log_file_path,
		# display_name='File Log Path',
		# )

		## Console Logging
		use_log_console: bool = BLField(
			gather.blext_info(blext_pkg).use_log_console,
			display_name='Log to Console',
		)
		# log_console_level: bl_logger.BLExtLogLevel = BLField(
		# gather.blext_info(blext_pkg).log_console_level,
		# display_name='Console Log Level',
		# )

		####################
		# - Logging
		####################
		def conform_logger(self, logger: logging.Logger) -> None:
			"""Conform a logger's settings to the extension preferences.

			Parameters:
				logger: The logger to configure using settings in the addon preferences.
			"""
			bl_logger.update_logger(
				logger,
				cb_console_handler=bl_logger.sensible_console_handler,
				cb_file_handler=bl_logger.sensible_file_handler,
				console_level=(
					self.log_console_level.log_level if self.use_log_console else None
				),
				file_path=(self.log_file_path if self.use_log_file else None),
				file_level=(
					self.log_file_level.log_level if self.use_log_file else None
				),
			)

		# @bl_event.run_on_event(bl_event.Event.BLClassRegistered)
		# @bl_event.run_on_event(
		# bl_event.Event.PropChanged,
		# on_props={
		# 'use_log_file',
		# 'log_file_level',
		# 'log_file_path',
		# 'use_log_console',
		# 'log_console_level',
		# },
		# )
		def conform_all_loggers_to_prefs(self) -> None:
			"""Called to reconfigure all loggers to match newly-altered addon preferences.

			This causes ex. changes to desired console log level to immediately be applied, but only the this addon's loggers.

			Parameters:
				single_logger_to_setup: When set, only this logger will be setup.
					Otherwise, **all addon loggers will be setup**.
			"""
			blext_pkg = self.bl_idname
			for logger in bl_logger.all_blext_loggers(blext_pkg):
				self.setup_logger(logger)

		####################
		# - UI
		####################
		def draw_logging_prefs(self, layout: bpy.types.UILayout) -> None:
			"""Draw the standardized addon preferences into a UILayout.

			Examples:
				Run this within the `draw()` method of the addon preferences class like so:
				```python
				layout = self.layout
				self.draw_standard_prefs(layout)
				```

			Parameters:
				context: The Blender context object.
			"""
			####################
			# - Logging
			####################
			# Box: Log Level
			box = layout.box()
			row = box.row()
			row.alignment = 'CENTER'
			row.label(text='Logging')

			# Split
			split = box.split(factor=0.5)

			## Split Col: Console Logging
			col = split.column()
			row = col.row()
			self.draw_blfield(row, 'use_log_console', toggle=True)

			row = col.row()
			row.enabled = self.use_log_console
			# self.draw_blfield(row, 'log_level_console')

			## Split Col: File Logging
			col = split.column()
			row = col.row()
			self.draw_blfield(row, 'use_log_file', toggle=True)

			row = col.row()
			row.enabled = self.use_log_file
			# self.draw_blfield(row, 'log_file_path')

			row = col.row()
			row.enabled = self.use_log_file
			# self.draw_blfield(row, 'log_level_file')

	return _LoggingPrefs


def SensiblePrefs(blext_pkg: str) -> type[type]:
	class _SensiblePrefs(LoggingPrefs(blext_pkg)):
		"""Sensible set of baseline preferences fields, for use in extensions.

		Implements the following `blroots` integration:

		- `bl_logger`: Enable the integration of `blroots` logging with extension preferences.
		"""

	return _SensiblePrefs
