# blroots
# Copyright (C) blroots blext Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.	If not, see <http://www.gnu.org/licenses/>.

"""Manages the registration of Blender classes, including delayed registrations that require access to Python dependencies.

Attributes:
	_REGISTERED_CLASSES: Currently registered Blender classes.
"""

from collections import defaultdict

import bpy

from blroots import bl_prop
from blroots import typ as bltyp
from blroots.utils import bl_logger

log = bl_logger.get(__name__)

####################
# - Globals
####################
_REGISTERED_CLASSES: defaultdict[bltyp.BLExtName, list[bltyp.BLClass]] = defaultdict(
	list
)


####################
# - Class Registration
####################
def register_classes(
	blext_pkg: bltyp.BLExtName,
	bl_classes: list[bltyp.BLClass],
) -> None:
	"""Registers a list of Blender classes, on behalf of a named extension.

	Parameters:
		blext_pkg: The name of the extension to register Blender classes on behalf of.
		bl_classes: List of Blender classes to register.
	"""
	log.info('Registering %s Classes', len(bl_classes))
	for cls in bl_classes:
		# Skip if Already Registered
		if (
			hasattr(cls, 'bl_idname')
			and cls.bl_idname in _REGISTERED_CLASSES[blext_pkg]
		):
			msg = f'Skipping register of {cls.bl_idname}'
			log.info(msg)
			continue

		# Register BLIDProps
		## - These have a .register() method already.
		if issubclass(cls, bl_prop.BLIDProps):
			log.debug(
				'Registering BLIDProps: %s',
				repr(cls),
			)
			cls.register()

		# Register BLClass
		else:
			log.debug(
				'Registering Class: %s',
				repr(cls),
			)
			bpy.utils.register_class(cls)

		_REGISTERED_CLASSES[blext_pkg].append(cls)


def unregister_classes(blext_pkg: bltyp.BLExtName) -> None:
	"""Unregisters all previously registered Blender classes for an extension.

	Parameters:
		blext_pkg: The name of the extension to unregister Blender classes on behalf of.
	"""

	log.info('Unregistering %s Classes', len(_REGISTERED_CLASSES[blext_pkg]))
	for cls in reversed(_REGISTERED_CLASSES[blext_pkg]):
		# Unregister BLIDProps
		## - These have a .unregister() method already.
		if issubclass(cls, bl_prop.BLIDProps):
			log.debug(
				'Unregistering BLIDProps: %s',
				repr(cls),
			)
			cls.unregister()

		# Unregister BLClass
		else:
			log.debug(
				'Unregistering Class %s',
				repr(cls),
			)
			bpy.utils.unregister_class(cls)

	_REGISTERED_CLASSES[blext_pkg].clear()
