# blroots
# Copyright (C) blroots blext Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.	If not, see <http://www.gnu.org/licenses/>.

"""Manages the registration of Blender keymaps.

Attributes:
	_REGISTERED_KEYMAPS: Addon keymaps currently registered by this addon.
		Each addon keymap is constrained to a single `space_type`, which is the key.
	_REGISTERED_KEYMAP_ITEMS: Addon keymap items currently registered by this addon.
		Each keymap item is paired to the keymap within which it is registered.
		_Each keymap is guaranteed to also be found in `_REGISTERED_KEYMAPS`._
"""

from collections import defaultdict

import bpy

from blroots import typ as bltyp
from blroots.utils import bl_logger
from blroots.utils.bl_keymap_item import BLKeymapItem

log = bl_logger.get(__name__)

####################
# - Globals
####################
_REGISTERED_KEYMAPS: defaultdict[
	bltyp.BLExtName, dict[bltyp.BLSpaceType, bpy.types.KeyMap]
] = defaultdict(dict)

_REGISTERED_KEYMAP_ITEMS: defaultdict[
	bltyp.BLExtName, list[tuple[bpy.types.KeyMap, bpy.types.KeyMapItem]]
] = defaultdict(list)


####################
# - Keymap Registration
####################
def register_keymaps(
	blext_pkg: bltyp.BLExtName,
	keymap_items: list[BLKeymapItem],
) -> None:
	"""Registers a list of Blender hotkey definitions.

	Parameters:
		blext_pkg: The name of the extension to register Blender keymap items for.
		bl_keymap_items: List of Blender hotkey definitions to register.
	"""
	# Aggregate Requested Spaces of All Keymap Items
	keymap_space_types: set[bltyp.BLSpaceType] = {
		keymap_item.space_type for keymap_item in keymap_items
	}

	# Create Addon Keymap per Requested Space
	for keymap_space_type in keymap_space_types:
		log.info('Registering %s Keymap', keymap_space_type)
		bl_keymap = bpy.context.window_manager.keyconfigs.addon.keymaps.new(
			name=f'{blext_pkg} - {keymap_space_type}',
			space_type=keymap_space_type,
		)
		_REGISTERED_KEYMAPS[blext_pkg][keymap_space_type] = bl_keymap

	# Register Keymap Items in Correct Addon Keymap
	for keymap_item in keymap_items:
		log.info('Registering %s Keymap Item', keymap_item)
		bl_keymap = _REGISTERED_KEYMAPS[blext_pkg][keymap_item.space_type]
		bl_keymap_item = keymap_item.register(bl_keymap)

		_REGISTERED_KEYMAP_ITEMS[blext_pkg].append((bl_keymap, bl_keymap_item))


def unregister_keymaps(blext_pkg: bltyp.BLExtName) -> None:
	"""Unregisters all Blender keymaps associated with the addon.

	Parameters:
		blext_pkg: The name of the extension to register Blender keymap items for.
	"""
	# Unregister Keymap Items from Correct Addon Keymap
	for bl_keymap, bl_keymap_item in reversed(_REGISTERED_KEYMAP_ITEMS[blext_pkg]):
		log.info('Unregistering %s BL Keymap Item', bl_keymap_item)
		bl_keymap.keymap_items.remove(bl_keymap_item)

	_REGISTERED_KEYMAP_ITEMS[blext_pkg].clear()

	# Delete Addon Keymaps
	for bl_keymap in reversed(_REGISTERED_KEYMAPS[blext_pkg].values()):
		log.info('Unregistering %s Keymap', bl_keymap)
		bpy.context.window_manager.keyconfigs.addon.keymaps.remove(bl_keymap)

	_REGISTERED_KEYMAPS[blext_pkg].clear()
