# blroots
# Copyright (C) 2025 blroots Project Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements various key caches on instances of Blender objects, especially nodes and sockets."""

import typing as typ
import uuid

import bpy

from blroots.bl_prop.signal import Signal

InstanceID: typ.TypeAlias = str
SupportedBLCacheType: typ.TypeAlias = (
	bpy.types.ID | bpy.types.Operator | bpy.types.AddonPreferences
)

####################
# - Global Variables
####################
_CACHE_NONPERSIST: dict[InstanceID, dict[str, typ.Any]] = {}


####################
# - Create/Invalidate
####################
def _bl_instance_nonpersist_cache(
	bl_instance: SupportedBLCacheType,
) -> dict[str, typ.Any]:
	"""Retrieve the non-persistent cache of a BLInstance."""
	# Create Non-Persistent Cache Entry
	## Prefer explicit cache management to 'defaultdict'
	if _CACHE_NONPERSIST.get(bl_instance.instance_id) is None:
		_CACHE_NONPERSIST[bl_instance.instance_id] = {}

	return _CACHE_NONPERSIST[bl_instance.instance_id]


def invalidate_nonpersist_instance_id(instance_id: InstanceID) -> None:
	"""Invalidate any `instance_id` that might be utilizing cache space in `_CACHE_NONPERSIST`.

	Notes:
		This should be run by the `instance_id` owner in its `free()` method.

	Parameters:
		instance_id: The ID of the Blender object instance that's being freed.
	"""
	_CACHE_NONPERSIST.pop(instance_id, None)


####################
# - Access
####################
def read(
	bl_instance: SupportedBLCacheType | None,
	*,
	key: str,
	use_nonpersist: bool,
	use_persist: bool,
) -> typ.Any | typ.Literal[Signal.CacheNotReady, Signal.CacheEmpty]:
	"""Read the cache associated with a Blender Instance, without writing to it.

	Attributes:
		key: The name to read from the instance-specific cache.
		use_nonpersist: If true, will always try the non-persistent cache first.
		use_persist: If true, will always try accessing the attribute `bl_instance.<key>`, where `key` is the value of the same-named parameter.
			Generally, such an attribute should be a `bpy.types.Property`.

	Return:
		The cache hit, if any; else `Signal.CacheEmpty`.
	"""
	# Check BLInstance Readiness
	## - 'instance_id' should be initialized beforehand, somewhere else.
	if bl_instance is None:
		return Signal.CacheNotReady

	if not hasattr(bl_instance, 'instance_id') or bl_instance.instance_id == '':
		bl_instance.reset_instance_id()

	# Use Non-Persist: Try Hit on Non-Persistent Cache
	## - If empty, allow fallback to the persistent cache if requested.
	if use_nonpersist:
		cache_nonpersist = _bl_instance_nonpersist_cache(bl_instance)
		value = cache_nonpersist.get(key, Signal.CacheEmpty)
		if value is not Signal.CacheEmpty or not use_persist:
			return value

	# Use Persist: Try Hit on Persistent Cache
	## - Specifying 'use_persist' will force use of the persistent cache.
	if use_persist:
		return getattr(bl_instance, key, Signal.CacheEmpty)

	msg = "Must specify either 'use_nonpersist' or 'use_persist'."
	raise ValueError(msg)


def write(
	bl_instance: SupportedBLCacheType,
	*,
	key: str,
	value: typ.Any,  ## TODO: "Serializable" type
	use_nonpersist: bool,
	use_persist: bool,
) -> typ.Literal[Signal.CacheNotReady] | None:
	"""Write to the cache associated with a Blender Instance.

	Attributes:
		key: The name to write to the instance-specific cache.
		use_nonpersist: If true, will always write to the non-persistent cache first.
		use_persist: If true, will always write to attribute `bl_instance.key`, where `key` is the value of the same-named parameter.
			Generally, such an attribute should be a `bpy.types.Property`.
		call_on_prop_changed: Whether to trigger `bl_instance.on_prop_changed()` with the
	"""
	# Check BLInstance Readiness
	## - 'instance_id' should be initialized beforehand, somewhere else.
	if bl_instance is None:
		return Signal.CacheNotReady

	if not hasattr(bl_instance, 'instance_id') or bl_instance.instance_id == '':
		bl_instance.reset_instance_id()

	# Use Non-Persist: Write Non-Persistent Cache
	if use_nonpersist:
		cache_nonpersist = _bl_instance_nonpersist_cache(bl_instance)
		cache_nonpersist[key] = value

		return None

	# Use Persist: Try Hit on Persistent Cache
	## - Specifying 'use_persist' will force use of the persistent cache.
	if use_persist:
		setattr(bl_instance, key, value)

		return None

	msg = "Must specify either 'use_nonpersist' or 'use_persist'."
	raise ValueError(msg)


def invalidate_nonpersist(
	bl_instance: SupportedBLCacheType,
	key: str,
) -> typ.Literal[Signal.CacheNotReady] | None:
	"""Invalidate a particular key of the non-persistent cache.

	**Persistent caches can't be invalidated without writing to them**.
	To get the same effect, consider using `write()` to write its default value (which must be manually tracked).
	"""
	# Check BLInstance Readiness
	## - 'instance_id' should be initialized beforehand, somewhere else.
	if (
		bl_instance is None
		or not hasattr(bl_instance, 'instance_id')
		or bl_instance.instance_id == ''
	):
		return Signal.CacheNotReady

	# Invalidate Non-Persistent Cache Key
	cache_nonpersist = _bl_instance_nonpersist_cache(bl_instance)
	cache_nonpersist.pop(key, None)

	return None
